﻿#include "DriverSettingPage.h"
#include "InstalledDriverDialog.h"
#include "core/SourceSetting.h"
#include "mylib/wtlext/folderdialog.h"
#include "core/TemporaryDirectory.h"
#include "mylib/cabcompressor.h"
#include "mylib/cabdecompressor.h"
#include "mylib/FileCompressionInfo.h"


namespace fs {
using namespace boost::filesystem;
};

class TemporaryDirectoryManager
{
	TemporaryDirectoryCloseDelete temp_dir_close_del;
	std::vector<TemporaryDirectory*> temp_dirs;

public:
	~TemporaryDirectoryManager()
	{
		std::vector<TemporaryDirectory*>::iterator it;
		for (it = temp_dirs.begin(); it != temp_dirs.end(); ++it) {
			delete *it;
		}
	}

	fs::wpath createTempDir()
	{
		TemporaryDirectory *ptr = new TemporaryDirectory(&(this->temp_dir_close_del));
		temp_dirs.push_back(ptr);
		return ptr->getDirectoryPath();
	}
};

/////////////////////////////////////////////////////////////////////////////
// デバイスドライバ リスト ファイル関係
/////////////////////////////////////////////////////////////////////////////

#pragma pack(push, 1)
struct DRIVER_LIST_FILE_HEADER {
	char signature[16];
	uint64_t creation_time;
	uint32_t os_ver;
	int32_t product_type;
	int32_t architecture_id;
	uint32_t desc_address;
	uint32_t desc_size;
	uint32_t data_address;
	uint32_t data_size;
};
#pragma pack(pop)

class DriverListFileView :
	public CDialogImpl<DriverListFileView>
{
	CEdit edit;

public:
	enum { IDD = IDD_DRIVER_LIST_FILE_VIEW };

	void setText(std::wstring const &str)
	{
		this->edit.SetWindowText(str.c_str());
	}

	BEGIN_MSG_MAP_EX(DriverListFileView)
		MSG_WM_INITDIALOG(OnInitDialog)
	END_MSG_MAP_EX()

	BOOL OnInitDialog(HWND /*wndFocus*/, LPARAM /*lInitParam*/)
	{
		CMenuHandle sys_menu = this->GetSystemMenu(false);
		sys_menu.EnableMenuItem(SC_CLOSE, MF_GRAYED);
		this->edit = this->GetDlgItem(IDC_EDIT1);
		return true;
	}
};

class DriverListFileDialog :
	public CFileDialogImpl<DriverListFileDialog>
{
	DriverListFileView dlg;

public:
	DriverListFileDialog(BOOL bOpenFileDialog, // TRUE for FileOpen, FALSE for FileSaveAs
		LPCTSTR lpszDefExt = NULL,
		LPCTSTR lpszFileName = NULL,
		DWORD dwFlags = OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
		LPCTSTR lpszFilter = NULL,
		HWND hWndParent = NULL)
		: CFileDialogImpl<DriverListFileDialog>(bOpenFileDialog, lpszDefExt, lpszFileName, dwFlags, lpszFilter, hWndParent)
	{ }

	//BEGIN_MSG_MAP_EX(DriverListFileDialog)
	//	MSG_WM_INITDIALOG(OnInitDialog)
	//END_MSG_MAP_EX()

	//BOOL OnInitDialog(HWND /*wndFocus*/, LPARAM /*lInitParam*/)
	//{
	//	return true;
	//}

	void OnInitDone(LPOFNOTIFY /*lpon*/)
	{
		CRect rect;
		this->GetParent().GetWindowRect(rect);
		dlg.Create(this->m_hWnd);
		dlg.SetWindowPos(NULL, rect.right + 1, rect.top, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
		dlg.ShowWindow(SW_SHOW);
	}

	//void OnFolderChange(LPOFNOTIFY /*lpon*/)
	//{
	//	//dlg.setText(L"");
	//}

	void OnSelChange(LPOFNOTIFY /*lpon*/)
	{
		wchar_t buf[MAX_PATH];
		this->GetFilePath(buf, MAX_PATH);
		//dlg.setText(buf);
		// 
		fs::wpath const file_path = buf;
		if (!fs::exists(file_path) || !fs::is_regular_file(file_path)) {
			dlg.setText(L"");
			return;
		}

		boost::uintmax_t const file_size = fs::file_size(file_path);
		if (file_size <= sizeof(DRIVER_LIST_FILE_HEADER)) {
			dlg.setText(L"");
			return;
		}
		fs::ifstream ifs(file_path, std::ios::in | std::ios::binary);
		if (!ifs) {
			dlg.setText(L"");
			return;
		}
		boost::scoped_array<char> header_buf(new char[sizeof(DRIVER_LIST_FILE_HEADER)]);
		ifs.read(header_buf.get(), sizeof(DRIVER_LIST_FILE_HEADER));
		DRIVER_LIST_FILE_HEADER const *header_ptr = reinterpret_cast<DRIVER_LIST_FILE_HEADER*>(header_buf.get());
		if (::memcmp(header_ptr->signature, "F6ToolDriverList", 16) != 0) {
			dlg.setText(L"");
			return;
		}
		std::wstringstream edit_ctrl;
		time_t const t0 = static_cast<time_t>(header_ptr->creation_time);
		edit_ctrl << L"作成日時 : ";
		// 日付
		{
			std::wstringstream ss;
			ss.imbue(std::locale(""));
			tm t = {0};
			::localtime_s(&t, &t0);
			const std::time_put<wchar_t>& tp =
				std::use_facet<std::time_put<wchar_t> >(ss.getloc());
			wchar_t const *fmt = L"%x %X";
			tp.put(ss, ss, ss.fill(),
				&t, fmt, fmt + 5);
			edit_ctrl << ss.str() << L"\r\n";
		}
		uint32_t const os_ver = header_ptr->os_ver;
		int const product_type = header_ptr->product_type;
		int const architecture_id = header_ptr->architecture_id;
		// from ConfirmationPage
		switch (os_ver) {
		case 0x00050000:
			edit_ctrl << _T("Windows 2000");
			if (product_type == 0) {
				edit_ctrl << L" Professional";
			} else if (product_type == 1) {
				edit_ctrl << L" Server";
			}
			break;
		case 0x00050001:
			edit_ctrl << _T("Windows XP");
			if (product_type == 0) {
				edit_ctrl << L" Professional";
			} else if (product_type == 4) {
				edit_ctrl << L" Home Edition";
			}
			break;
		case 0x00050002:
			if (product_type == 0) {
				edit_ctrl << L"Windows XP Professional";
			} else {
				edit_ctrl << _T("Windows Server 2003");
				if (product_type == 1) {
					edit_ctrl << L" Standard Edition";
				} else if (product_type == 2) {
					edit_ctrl << L" Enterprise Edition";
				} else if (product_type == 3) {
					edit_ctrl << L" Datacenter Edition";
				}
			}
			break;
		default:
			edit_ctrl << _T("不明なOS");
			break;
		}
		switch (architecture_id) {
		case PROCESSOR_ARCHITECTURE_AMD64:
			edit_ctrl << _T(" x64");
			break;
		case PROCESSOR_ARCHITECTURE_IA64:
			edit_ctrl << _T(" for Itanium-based systems");
			break;
		}
		edit_ctrl << L"\r\n\r\n";

		// driver descriptions
		size_t const desc_buf_size = header_ptr->desc_size;
		boost::scoped_array<char> desc_buf(new char[desc_buf_size]);
		ifs.seekg(header_ptr->desc_address, std::ios::beg);
		ifs.read(desc_buf.get(), desc_buf_size);
		wchar_t const * const desc_ptr_start = reinterpret_cast<wchar_t*>(desc_buf.get());
		wchar_t const *desc_ptr = desc_ptr_start;
		size_t desc_len = wcslen(desc_ptr);
		while (desc_len > 0 && (desc_ptr - desc_ptr_start + desc_len + 1)*sizeof(wchar_t) < desc_buf_size) {
			edit_ctrl << desc_ptr << L"\r\n";
			desc_ptr += desc_len + 1;
			desc_len = wcslen(desc_ptr);
		}

		dlg.setText(edit_ctrl.str());
	}
};

/////////////////////////////////////////////////////////////////////////////

DriverSettingPage::DriverSettingPage()
: architecture_id(PROCESSOR_ARCHITECTURE_INTEL/*-1*/)
, os_version(0x00050001)
{
	this->drivers_list.setParameter(architecture_id, os_version);
}

namespace {
int prod_type = 0;

void Fn1(DriverSetupData const &data,
         fs::wpath const &drivers_dir_path,
         size_t drivers_dir_path_len,
         DriverFileDataMap *f,
         DriverServiceData *s)
{
	fs::wpath const inf_path = data.getInfFilePath();
	fs::wpath tmp = inf_path.parent_path();
	while (tmp.string().length() >= drivers_dir_path_len) {
		std::wstring const sub_dir_name = tmp.filename();
		tmp = tmp.parent_path();
		if (fs::equivalent(tmp, drivers_dir_path)) {
			fs::wpath const drivers_sub_dir_path = drivers_dir_path/sub_dir_name;
			DriverFileDataMap::iterator it1 = f->find(sub_dir_name);
			if (it1 != f->end()) {
				it1->second.addFiles(data, drivers_sub_dir_path);
			} else {
				DriverFileDataMap::value_type v(sub_dir_name, DriverFileData());
				v.second.addFiles(data, drivers_sub_dir_path);
				f->insert(v);
			}
		}
	}

	s->addService(data);
}

// TODO: Integrator.cppとの重複解消
boost::filesystem::wpath changeExtensionForCompressedFile(boost::filesystem::wpath const &file_path, wchar_t tail_char)
{
	std::wstring const path_str = file_path.string();
	std::wstring const ext = fs::extension(file_path);

	if (ext.empty()) {
		// 拡張子なし
		wchar_t new_ext[] = {L'.', tail_char, L'\0'};
		return fs::change_extension(file_path, new_ext);
	} else {
		size_t const ext_len = ext.length();
		if (ext_len <= 3) {
			// 拡張子がドットを入れて3文字以下
			return fs::change_extension(file_path, ext + tail_char);
		} else {
			std::wstring new_ext = ext;
			new_ext[ext_len - 1] = tail_char;
			return fs::change_extension(file_path, new_ext);
		}
	}
}
// TODO: Integrator.cppとの重複解消
// copyFileWithCompress
//  source_file_path and dest_file_path have no-compressed filename.
void copyFileWithCompress(boost::filesystem::wpath const &source_file_path,
                          boost::filesystem::wpath const &dest_file_path,
                          std::wstring const &file_title)
{
	fs::wpath const dest_path =
		changeExtensionForCompressedFile(dest_file_path, L'_');

	MyLib::FileCompressionInfo info(source_file_path);
	if (info.getCompressionType() != FILE_COMPRESSION_NONE) {
		// ファイルが圧縮されていればそのままコピー
		std::wstring const src_path_str = info.getActualPath().string();
		std::wstring const dest_path_str = dest_path.string();
		MyLib::toFileWritable(dest_path);
		if (!::CopyFile(src_path_str.c_str(), dest_path_str.c_str(), FALSE)) {
			throw MY_WIN32_FILE_TRANSFER_EXCEPTION_LAST_ERROR(src_path_str, dest_path_str);
		}
	} else {
		// ファイルが圧縮されていなければ圧縮する
		MyLib::CabCompressor compress(dest_path);
		compress.addFile(source_file_path, file_title);
		compress.flush(false);
	}
}

bool is_same_architecture;
};

DriverSetting* DriverSettingPage::getSetting()
{
	fs::wpath const drivers_dir_path = this->drivers_dir_path_;
	size_t const drivers_dir_path_len = drivers_dir_path.string().length();

	std::auto_ptr<DriverSetting> result(new DriverSetting());
	int const item_count = this->drivers_list.GetItemCount();
	for (int i = 0; i < item_count; ++i) {
		ItemAssociatedData *ptr = reinterpret_cast<ItemAssociatedData*>(this->drivers_list.GetItemData(i));
		ItemAssociatedData::ItemState const item_state = ptr->getItemState();
		DriverSetupData const *item_data_ptr = ptr->getItemData();
		switch (item_state) {
		case ItemAssociatedData::ADDED_BEFORE:
			Fn1(*item_data_ptr, drivers_dir_path, drivers_dir_path_len, &(result->leave_drivers), &(result->leave_services));
			break;
		case ItemAssociatedData::ADDED_BEFORE_REMOVE:
			Fn1(*item_data_ptr, drivers_dir_path, drivers_dir_path_len, &(result->del_drivers), &(result->del_services));
			break;
		case ItemAssociatedData::NEWLY_ADDED:
			result->add_drivers.push_back(*item_data_ptr);
			break;
		default:
			break;
		}
	}
	return result.release();
}

void DriverSettingPage::loadSourceSetting(SourceSetting const *source_setting)
{
	prod_type = source_setting->getProductType();
	this->architecture_id = source_setting->getArchitectureID();
	this->os_version = source_setting->getOSVersion();
	this->drivers_list.DeleteAllItems();
	this->drivers_list.setParameter(architecture_id, os_version);

	type_info const &t_info = typeid(*source_setting);
	if (t_info == typeid(DirectorySourceSetting) || t_info == typeid(DiscSourceSetting)) {
		DirectorySourceSetting const *ptr = static_cast<DirectorySourceSetting const*>(source_setting);
		fs::wpath const src_path = ptr->getSourcePath();
		fs::wpath const drivers_path = src_path/L"$OEM$/Drivers";
		if (fs::exists(drivers_path) && fs::is_directory(drivers_path)) {
			this->drivers_list.addIntegratedDrivers(drivers_path);
		}
		this->drivers_dir_path_ = drivers_path;
	}

	// 「インストールされているドライバの追加」の有効・無効切り替え
	// F6回避ツールとターゲットOSで CPU Architecture が一致している必要あり
	typedef BOOL (WINAPI* LPFN_ISWOW64PROCESS)(HANDLE hProcess, PBOOL Wow64Process);
	static LPFN_ISWOW64PROCESS IsWow64Process = (LPFN_ISWOW64PROCESS)::GetProcAddress(
		::GetModuleHandle(L"kernel32.dll"), "IsWow64Process");
	is_same_architecture = true;
	if (IsWow64Process != NULL) {
		BOOL is_wow64_process;
		IsWow64Process(::GetCurrentProcess(), &is_wow64_process);
		is_same_architecture = !is_wow64_process;
	}
	//this->installed_driver_button.EnableWindow(is_same_architecture);
}


DWORD DriverSettingPage::OnPrePaint(int /*nID*/, NMCUSTOMDRAW *lpnmcd)
{
	if (lpnmcd->hdr.idFrom == IDC_LIST1) {
		return CDRF_NOTIFYITEMDRAW;
	} else {
		return CDRF_DODEFAULT;
	}
}

DWORD DriverSettingPage::OnItemPrePaint(int /*nID*/, NMCUSTOMDRAW *lpnmcd)
{
	if(lpnmcd->hdr.idFrom == IDC_LIST1){
		return CDRF_NOTIFYPOSTPAINT;
	} else {
		return CDRF_DODEFAULT;
	}
}

DWORD DriverSettingPage::OnItemPostPaint(int /*nID*/, NMCUSTOMDRAW *lpnmcd)
{
	if(lpnmcd->hdr.idFrom == IDC_LIST1){
		//NMLVCUSTOMDRAW *lpnmlv = reinterpret_cast<NMLVCUSTOMDRAW*>(lpnmcd);
		int const item_index = lpnmcd->dwItemSpec;
		ItemAssociatedData *ptr =
			reinterpret_cast<ItemAssociatedData*>(this->drivers_list.GetItemData(item_index));
		if (ptr->getItemState() == ItemAssociatedData::ADDED_BEFORE_REMOVE) {
			//lpnmlv->clrText = RGB(255, 255, 255);
			//lpnmlv->clrTextBk = RGB(128, 128, 128);
			CRect r;
			this->drivers_list.GetItemRect(item_index, r, LVIR_BOUNDS);
			int const line_y = r.Height()/2;
			CPen pen;
			pen.CreatePen(PS_SOLID, 2, RGB(255, 0, 0));
			CDCHandle dc(lpnmcd->hdc);
			HPEN default_pen = dc.SelectPen(pen);
			dc.MoveTo(r.left, r.top + line_y);
			dc.LineTo(r.right , r.top + line_y);
			dc.SelectPen(default_pen);
		}
	}
	return CDRF_DODEFAULT;
}


BOOL DriverSettingPage::OnInitDialog(HWND /*wndFocus*/, LPARAM /*lInitParam*/)
{
#ifdef MY_DEBUG
	::WritePrivateProfileString(L"Debug", L"State", L"DriverSettingPage::OnInitDialog() start", PROFILE_PATH->string().c_str());
#endif
	//this->installed_driver_button = this->GetDlgItem(IDC_BUTTON_INSTALLED_DRIVER);
	this->add_button      = this->GetDlgItem(IDC_BUTTON_ADD);
	this->remove_button   = this->GetDlgItem(IDC_BUTTON_REMOVE);
	this->property_button = this->GetDlgItem(IDC_BUTTON_PROPERTY);

	this->drivers_list.SubclassWindow(this->GetDlgItem(IDC_LIST1));
	this->drivers_list.SetExtendedListViewStyle(LVS_EX_INFOTIP);
	this->drivers_list.DragAcceptFiles();
	this->drivers_list.setDeviceClassImageList();

	CRect rect;
	this->drivers_list.GetClientRect(&rect);
	const int list_view_width = rect.Width() - 10;
	const int col2_width = list_view_width/5;
	const int col1_width = col2_width*2;
	const int col3_width = list_view_width - col1_width - col2_width;
	this->drivers_list.InsertColumn(0, L"名前", 0, col1_width);
	this->drivers_list.SetColumnSortType(0, LVCOLSORT_TEXTNOCASE);
	this->drivers_list.InsertColumn(1, L"種類", 0, col2_width);
	this->drivers_list.SetColumnSortType(1, LVCOLSORT_TEXTNOCASE);
	this->drivers_list.InsertColumn(2, L"ドライバの場所", 0, col3_width);
	this->drivers_list.SetColumnSortType(2, LVCOLSORT_TEXTNOCASE);

	this->DlgResize_Init(false);
#ifdef MY_DEBUG
	::WritePrivateProfileString(L"Debug", L"State", L"DriverSettingPage::OnInitDialog() end", PROFILE_PATH->string().c_str());
#endif
	return TRUE;
}

void DriverSettingPage::OnContextMenu(CWindow wnd, CPoint pt)
{
	// [Shift]+[F10]キーが押された場合は座標をクライアント領域の左上に設定
	if(pt.x == -1 && pt.y == -1){
		pt.SetPoint(0, 0);
		wnd.ClientToScreen(&pt);
	}
	// 座標がクライアント領域内の場合のみポップアップメニューを表示
	CRect rc;
	wnd.GetClientRect(&rc);
	wnd.ClientToScreen(reinterpret_cast<POINT*>(&rc) + 1);
	if (rc.PtInRect(pt)){
		CMenu menu_popup;
		menu_popup.LoadMenu(IDR_MENU_DRIVERS_LIST);
		menu_popup.GetSubMenu(0).TrackPopupMenu(
			TPM_LEFTALIGN | TPM_TOPALIGN | TPM_LEFTBUTTON, pt.x, pt.y, this->m_hWnd);
	}
}

LRESULT DriverSettingPage::OnMenuSelect(unsigned int /*uMsg*/, WPARAM wParam, LPARAM lParam)
{
	return this->GetParent().PostMessage(UWM_POPUP_MENU_SELECT, wParam, lParam);
}

void DriverSettingPage::OnButtonAdd(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	CMenu menu;
	menu.LoadMenu(IDR_MENU_ADD_DRIVERS);
	CMenuHandle sub_menu = menu.GetSubMenu(0);
	if (!is_same_architecture) {
		sub_menu.EnableMenuItem(IDM_ADD_DRIVERS_INSTALLED, MF_GRAYED);
	}
	CRect rect;
	this->add_button.GetWindowRect(rect);
	sub_menu.TrackPopupMenu(TPM_LEFTALIGN | TPM_LEFTBUTTON,
	                        rect.left,
	                        rect.bottom - 1,
	                        this->m_hWnd,
	                        rect);
}

//void DriverSettingPage::OnButtonFile(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
void DriverSettingPage::OnAddDriversInf(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	OpenFileDialog dlg(NULL, NULL, OFN_HIDEREADONLY | OFN_FILEMUSTEXIST,
		_T("INFファイル (*.inf)\0*.inf\0"));
	if (dlg.DoModal() == IDOK) {
		CWaitCursor cursor;
		DriversListViewCtrl::ItemAdditionCount count =
			this->drivers_list.addDriversFromFileOrDir(fs::wpath(dlg.m_szFileName));
		this->drivers_list.showItemAdditionMessage(count);
	}
}

//void DriverSettingPage::OnButtonFolder(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
void DriverSettingPage::OnAddDriversFolder(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	// フォルダの参照
	static std::wstring initial_dir_path;
	MyLib::FolderDialog dlg(initial_dir_path, this->m_hWnd, NULL,
		BIF_RETURNONLYFSDIRS | BIF_NEWDIALOGSTYLE | BIF_NONEWFOLDERBUTTON);
	if (dlg.DoModal() == IDOK) {
		CWaitCursor cursor;
		DriversListViewCtrl::ItemAdditionCount count =
			this->drivers_list.addDriversFromFileOrDir(fs::wpath(dlg.GetFolderPath()));
		this->drivers_list.showItemAdditionMessage(count);
		initial_dir_path = dlg.GetFolderPath();
	}
}

//void DriverSettingPage::OnButtonInstalledDriver(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
void DriverSettingPage::OnAddDriversInstalled(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	//int add_count = 0;
	InstalledDriverDialog dlg(this->architecture_id, this->os_version);
	if (dlg.DoModal() == IDOK) {
		CWaitCursor cursor;
		DriversListViewCtrl::ItemAdditionCount count =
			this->drivers_list.addDriversListItems(dlg.getData(), L"<システムにインストールされているデバイスドライバ>");
		this->drivers_list.showItemAdditionMessage(count);
	}
}

void DriverSettingPage::OnAddDriversImport(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	DriverListFileDialog dlg(true, NULL, NULL, OFN_HIDEREADONLY | OFN_FILEMUSTEXIST,
		L"ドライバ リスト ファイル (*.driver_list)\0*.driver_list\0");
	if (dlg.DoModal() != IDOK) {
		return;
	}
	try {
		CWaitCursor wait_cursor;

		fs::wpath const file_path = dlg.m_szFileName;
		boost::uintmax_t const file_size = fs::file_size(file_path);
		if (file_size <= sizeof(DRIVER_LIST_FILE_HEADER)) {
			this->MessageBox(L"未対応のファイル形式です");
			return;
		}
		fs::ifstream ifs(file_path, std::ios::in | std::ios::binary);
		boost::scoped_array<char> header_buf(new char[sizeof(DRIVER_LIST_FILE_HEADER)]);
		ifs.read(header_buf.get(), sizeof(DRIVER_LIST_FILE_HEADER));
		DRIVER_LIST_FILE_HEADER const *header_ptr = reinterpret_cast<DRIVER_LIST_FILE_HEADER*>(header_buf.get());
		uint32_t const data_address = header_ptr->data_address;
		size_t const data_size = header_ptr->data_size;
		if (::memcmp(header_ptr->signature, "F6ToolDriverList", 16) != 0
		 || file_size < data_address + data_size) {
			this->MessageBox(L"未対応のファイル形式です");
			return;
		}
		// -----------------------------
		TemporaryDirectoryCloseDelete temp_dir_close_del;
		TemporaryDirectory temp_dir(&temp_dir_close_del);
		fs::wpath const temp_dir_path = temp_dir.getDirectoryPath();
		// ドライバリストファイルからcabファイルデータを取り出す
		boost::scoped_array<char> data_buf(new char[data_size]);
		ifs.seekg(data_address, std::ios::beg);
		ifs.read(data_buf.get(), data_size);
		fs::wpath const cab_file_path = MyLib::createTemporaryFilePath(temp_dir_path);
		{
			fs::ofstream ofs(cab_file_path, std::ios::out | std::ios::binary);
			ofs.write(data_buf.get(), data_size);
		}
		// アプリケーションの終了まで解凍ファイルが削除されないようにする
		static TemporaryDirectoryManager temp_dir_mgr;
		fs::wpath const drivers_dir_path = temp_dir_mgr.createTempDir();
		MyLib::CabDecompressor decomp;
		if (!decomp.extractCabFile(cab_file_path, drivers_dir_path)) {
			this->MessageBox(L"ドライバの読み込みに失敗しました");
			return;
		}
		DriversListViewCtrl::ItemAdditionCount count =
			this->drivers_list.addDriversFromFileOrDir(drivers_dir_path, file_path.native_file_string().c_str());
		this->drivers_list.showItemAdditionMessage(count);

	} catch (MyLib::FileTransferException<wchar_t>& e) {
		std::wostringstream ss;
		ss << L"From: " << e.getSourceFileName() << L"\n";
		ss << L"To: " << e.getDestinationFileName() << L"\n";
		if (e.getErrorFile() != NULL) {
			ss << e.getErrorFile() << L"(" << e.getErrorLine() << L")";
		}
		ss << e.getErrorText();
		this->MessageBox(ss.str().c_str());
	} catch (MyLib::FileException<wchar_t>& e) {
		std::wostringstream ss;
		ss << e.getFileName() << L"\n";
		if (e.getErrorFile() != NULL) {
			ss << e.getErrorFile() << L"(" << e.getErrorLine() << L")";
		}
		ss << e.getErrorText();
		this->MessageBox(ss.str().c_str());
	} catch (MyLib::Exception<wchar_t>& e) {
		std::wostringstream ss;
		if (e.getErrorFile() != NULL) {
			ss << e.getErrorFile() << L"(" << e.getErrorLine() << L")";
		}
		ss << e.getErrorText() << L'[' << e.getErrorCode() << L']';
		this->MessageBox(ss.str().c_str());
	}
}

void DriverSettingPage::OnButtonExport(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	int const item_count = this->drivers_list.GetItemCount();
	bool deleted_only = true;
	for (int i = 0; i < item_count; ++i) {
		ItemAssociatedData *ptr = reinterpret_cast<ItemAssociatedData*>(this->drivers_list.GetItemData(i));
		ItemAssociatedData::ItemState const item_state = ptr->getItemState();
		if (item_state != ItemAssociatedData::ADDED_BEFORE_REMOVE) {
			deleted_only = false;
			break;
		}
	}
	if (deleted_only) {
		// 一覧が空
		this->MessageBox(L"デバイスドライバを追加してください");
		return;
	}

	DriverListFileDialog dlg(false, L"driver_list", NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
		L"ドライバ リスト ファイル (*.driver_list)\0*.driver_list\0");
	if (dlg.DoModal() != IDOK) {
		return;
	}

#ifdef NDEBUG
	std::wostringstream ss;
	try {
#endif
		CWaitCursor wait_cursor;

		// 一時ディレクトリ作成
		TemporaryDirectoryCloseDelete temp_dir_close_del;
		TemporaryDirectory temp_dir(&temp_dir_close_del);
		fs::wpath const temp_dir_path = temp_dir.getDirectoryPath();
		// ディレクトリ重複防止機構
		typedef std::map<std::wstring, boost::filesystem::wpath, MyLib::LessNoCase> DriverDestinations;
		DriverDestinations driver_dests;

		std::set<std::wstring, MyLib::LessNoCase> files_in_cab;

		size_t all_desc_buf_used = 0;
		std::vector<wchar_t> all_desc_buf;
		all_desc_buf.reserve(256*item_count);
		int desc_buf_size = 256;
		std::vector<wchar_t> desc_buf(desc_buf_size);

		fs::wpath const cab_file_path = MyLib::createTemporaryFilePath(temp_dir_path);
		MyLib::CabCompressor compress(cab_file_path);
		int dir_index = 0;
		for (int i = 0; i < item_count; ++i) {
			ItemAssociatedData *ptr = reinterpret_cast<ItemAssociatedData*>(this->drivers_list.GetItemData(i));
			ItemAssociatedData::ItemState const item_state = ptr->getItemState();
			if (item_state != ItemAssociatedData::ADDED_BEFORE_REMOVE) {
				// ドライバの説明
				int desc_len;
				for (;;) {
					desc_len = this->drivers_list.GetItemText(i, 0, &(desc_buf.at(0)), desc_buf_size);
					if (desc_len < desc_buf_size - 1) {
						break;
					}
					desc_buf_size *= 2;
					desc_buf.resize(desc_buf_size);
				}
				all_desc_buf.resize(all_desc_buf_used + desc_len + 1);
				memcpy(&(all_desc_buf.at(all_desc_buf_used)), &(desc_buf.at(0)), (desc_len + 1)*sizeof(wchar_t));
				all_desc_buf_used += desc_len + 1;

				DriverSetupData const *item_data_ptr = ptr->getItemData();
				DriverSetupData const &data = *item_data_ptr;

				// メインINFファイルのパスを取得
				fs::wpath const inf_src_path = data.getInfFilePath();

#ifdef NDEBUG
				try {
#endif
					// Installation Root
					fs::wpath installation_root = inf_src_path;
					installation_root.remove_filename();

					fs::wpath dest_dir_path;
					std::wstring const installation_root_str = installation_root.string();
					DriverDestinations::const_iterator dests_it = driver_dests.find(installation_root_str);
					if (dests_it != driver_dests.end()) {
						dest_dir_path = dests_it->second;
					} else {
						// 各ドライバに16進数の番号をつけて管理する
						TCHAR buf[4];
						::wsprintf(buf, _T("%X"), dir_index);
						dest_dir_path = buf;
						driver_dests[installation_root_str] = dest_dir_path;
						++dir_index;
					}

					// メインINFファイルの処理
					fs::wpath const inf_dest_path = dest_dir_path/data.getOriginalInfFileTitle();
					std::wstring const inf_dest_path_str = inf_dest_path.string();
					std::set<std::wstring, MyLib::LessNoCase>::const_iterator it001 = files_in_cab.find(inf_dest_path_str);
					if (it001 == files_in_cab.end()) {
						compress.addFile(inf_src_path, inf_dest_path_str);
						files_in_cab.insert(inf_dest_path_str);
					}
					// メインINFファイルのCopyInfsで指定されているファイルの処理
					DriverSetupData::CopyInfs::const_iterator it002 = data.getCopyInfs().begin();
					for (; it002 != data.getCopyInfs().end(); ++it002) {
						fs::wpath const src_path = installation_root/it002->second;
						fs::wpath const dest_path = dest_dir_path/it002->first;
						std::wstring const dest_path_str = dest_path.string();
						std::set<std::wstring, MyLib::LessNoCase>::const_iterator it01 = files_in_cab.find(dest_path_str);
						if (it01 == files_in_cab.end()) {
							compress.addFile(src_path, dest_path_str);
							files_in_cab.insert(dest_path_str);
						}
					}
					// NewCabFile 作成
					DriverSetupData::NewCabFileCollection const new_cab_files = data.getNewCabFiles();
					DriverSetupData::NewCabFileCollection::const_iterator it003 = new_cab_files.begin();
					for (; it003 != new_cab_files.end(); ++it003) {
						fs::wpath const dest_path = dest_dir_path/it003->first;
						std::wstring const dest_path_str = dest_path.string();
						std::set<std::wstring, MyLib::LessNoCase>::const_iterator it01 = files_in_cab.find(dest_path_str);
						if (it01 == files_in_cab.end()) {
							// cab file 作成
							fs::wpath const new_cab_file_path = MyLib::createTemporaryFilePath(temp_dir_path);
							{
							MyLib::CabCompressor compress1(new_cab_file_path);
							DriverSetupData::FileInCabInfo::const_iterator it37 = it003->second.begin();
							for (; it37 != it003->second.end(); ++it37) {
								compress1.addFile(it37->second, it37->first);
							}
							//compress1.flush(false);
							}
							// main cab file に追加
							compress.addFile(new_cab_file_path, dest_path_str);
							files_in_cab.insert(dest_path_str);
						}
					}
					// TagFileのコピー
					DriverSetupData::TagFileCollection const tag_files = data.getTagFiles();
					DriverSetupData::TagFileCollection::const_iterator it35 = tag_files.begin();
					for (; it35 != tag_files.end(); ++it35) {
						fs::wpath const src_path = installation_root/it35->first;
						fs::wpath const dest_path = dest_dir_path/it35->second;
						std::wstring const dest_path_str = dest_path.string();
						std::set<std::wstring, MyLib::LessNoCase>::const_iterator it01 = files_in_cab.find(dest_path_str);
						if (it01 == files_in_cab.end()) {
							if (fs::exists(src_path)) {
								compress.addFile(src_path, dest_path_str);
							} else {
								// 元のファイルが存在しない場合、とりあえず空のファイルを作成
								fs::wpath const tag_file_path = MyLib::createTemporaryFilePath(temp_dir_path);
								::CloseHandle(::CreateFile(tag_file_path.string().c_str(), GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL));
								compress.addFile(tag_file_path, dest_path_str);
							}
							files_in_cab.insert(dest_path_str);
						}
					}
					// メインINFファイルのCopyFilesで指定されているファイルの処理
					DriverSetupData::FileDataCollection::const_iterator it =
						data.getFileDataCollection().begin();
					for (; it != data.getFileDataCollection().end(); ++it) {
						DriverSetupData::CopyFilesTagData::const_iterator it2 =
							it->second.begin();
						for (; it2 != it->second.end(); ++it2) {
							std::wstring const file_title = it2->first;
							fs::wpath const source_file_path = it2->second.source_path;
							fs::wpath const dest_file_path = dest_dir_path/it2->second.installation_path;
							std::wstring const dest_file_path_str = dest_file_path.string();
							std::set<std::wstring, MyLib::LessNoCase>::const_iterator it01 = files_in_cab.find(dest_file_path_str);
							if (it01 == files_in_cab.end()) {
								// 拡張子をチェック
								std::wstring const ext = fs::extension(dest_file_path);
								// コピー
								if (MyLib::compareStringNoCase(ext, L".ini") != 0) {
									fs::wpath const dest_path =
										changeExtensionForCompressedFile(dest_file_path, L'_');
									std::wstring const dest_path_str = dest_path.string();
									MyLib::FileCompressionInfo info(source_file_path);
									fs::wpath const src_path = info.getActualPath();
									if (info.getCompressionType() != FILE_COMPRESSION_NONE) {
										// ファイルが圧縮されていればそのままコピー
										compress.addFile(src_path, dest_path_str);
									} else {
										fs::wpath const temp_file_path = MyLib::createTemporaryFilePath(temp_dir_path);
										// ファイルが圧縮されていなければ圧縮する
										{
											MyLib::CabCompressor compress1(temp_file_path);
											compress1.addFile(src_path, file_title);
										}
										// main cab file に追加
										compress.addFile(temp_file_path, dest_path_str);
									}
								} else {
									// iniファイルはそのままコピー
									compress.addFile(source_file_path, dest_file_path_str);
								}
								files_in_cab.insert(dest_file_path_str);
							}
						}
					}
					//
					// catファイルをコピー
					//
					fs::wpath const src_cat_file_path = data.getCatalogFilePath();
					if (fs::exists(src_cat_file_path)) {
						// 元のファイルが存在しなければスキップ
						// (存在しないファイルをINFファイルに書く糞メーカーがあるので。Intelとか、Intelとか、…)
						std::wstring const cat_file_title = data.getOriginalCatalogFileTitle();
						fs::wpath const dest_cat_file_path = dest_dir_path/cat_file_title;
						std::wstring const dest_path_str = dest_cat_file_path.string();
						std::set<std::wstring, MyLib::LessNoCase>::const_iterator it006 = files_in_cab.find(dest_path_str);
						if (it006 == files_in_cab.end()) {
							compress.addFile(src_cat_file_path, dest_path_str);
							files_in_cab.insert(dest_path_str);
						}
					}
#ifdef NDEBUG
				} catch (...) {
					ss << L'"' << &desc_buf.at(0) << L"\" 処理中のエラー\n";
					throw;
				}
#endif
			} // if (item_state != ItemAssociatedData::ADDED_BEFORE_REMOVE) {
		}
		compress.flush(false);
		all_desc_buf_used += 1;
		all_desc_buf.push_back(L'\0');

		fs::wpath const file_path = dlg.m_szFileName;
		fs::ofstream ofs(file_path, std::ios::out | std::ios::binary);
		// file header
		DRIVER_LIST_FILE_HEADER header;
		memcpy(header.signature, "F6ToolDriverList", 16);
		header.creation_time = static_cast<uint64_t>(std::time(NULL));
		header.os_ver = this->os_version;
		header.product_type = prod_type;
		header.architecture_id = this->architecture_id;
		header.desc_address = sizeof(DRIVER_LIST_FILE_HEADER);
		uint32_t const desc_size = static_cast<uint32_t>(all_desc_buf_used*sizeof(wchar_t));
		header.desc_size = desc_size;
		header.data_address = header.desc_address + header.desc_size;
		uint32_t const data_size = static_cast<uint32_t>(fs::file_size(cab_file_path));
		header.data_size = data_size;
		ofs.write(reinterpret_cast<char*>(&header), sizeof(DRIVER_LIST_FILE_HEADER));
		// driver descriptions
		ofs.write(reinterpret_cast<char*>(&all_desc_buf.at(0)), desc_size);
		// cab file data
		fs::ifstream ifs(cab_file_path, std::ios::in | std::ios::binary);
		boost::scoped_array<char> buf(new char[data_size]);
		ifs.read(buf.get(), data_size);
		ofs.write(buf.get(), data_size);

		this->MessageBox(L"ドライバ一覧の保存が完了しました");

#ifdef NDEBUG
	} catch (MyLib::FileTransferException<wchar_t>& e) {
		//std::wostringstream ss;
		ss << L"From: " << e.getSourceFileName() << L"\n";
		ss << L"To: " << e.getDestinationFileName() << L"\n";
		if (e.getErrorFile() != NULL) {
			ss << e.getErrorFile() << L"(" << e.getErrorLine() << L")";
		}
		ss << e.getErrorText();
		this->MessageBox(ss.str().c_str(), NULL, MB_ICONERROR);
	} catch (MyLib::FileException<wchar_t>& e) {
		//std::wostringstream ss;
		ss << e.getFileName() << L"\n";
		if (e.getErrorFile() != NULL) {
			ss << e.getErrorFile() << L"(" << e.getErrorLine() << L")";
		}
		ss << e.getErrorText();
		this->MessageBox(ss.str().c_str(), NULL, MB_ICONERROR);
	} catch (MyLib::Exception<wchar_t>& e) {
		//std::wostringstream ss;
		if (e.getErrorFile() != NULL) {
			ss << e.getErrorFile() << L"(" << e.getErrorLine() << L")";
		}
		ss << e.getErrorText() << L'[' << e.getErrorCode() << L']';
		this->MessageBox(ss.str().c_str(), NULL, MB_ICONERROR);
	} catch (const std::runtime_error& e) {
		std::wstring const err_msg = MyLib::convertUTF8ToUTF16(e.what());
		ss << err_msg;
		this->MessageBox(ss.str().c_str(), NULL, MB_ICONERROR);
	}
#endif
}

void DriverSettingPage::OnButtonRemove(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	this->drivers_list.removeSelectedItem();
}

void DriverSettingPage::OnButtonProperty(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	this->drivers_list.showPropertyOfSelectdeItem();
}

void DriverSettingPage::OnDriversListRemove(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	this->drivers_list.removeSelectedItem();
}

void DriverSettingPage::OnDriversListProperty(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	this->drivers_list.showPropertyOfSelectdeItem();
}

LRESULT DriverSettingPage::OnGetInfoTip(NMHDR *pnmh)
{
	NMLVGETINFOTIP *ptr = reinterpret_cast<NMLVGETINFOTIP*>(pnmh);
	static TCHAR buf[80];
	this->drivers_list.GetItemText(ptr->iItem, ptr->iSubItem, buf, 80);
	ptr->pszText = buf;

	return 0;
}

LRESULT DriverSettingPage::OnDeleteItem(NMHDR *pnmh)
{
	NMLISTVIEW *ptr = reinterpret_cast<NMLISTVIEW*>(pnmh);
	ItemAssociatedData *ptr2 =
		reinterpret_cast<ItemAssociatedData*>(this->drivers_list.GetItemData(ptr->iItem));
	delete ptr2;

	return 0;
}

LRESULT DriverSettingPage::OnItemChanged(NMHDR * /*pnmh*/)
{
	int i = this->drivers_list.GetNextItem(-1, LVIS_SELECTED);
	if (i == -1) {
		this->remove_button.EnableWindow(false);
		this->property_button.EnableWindow(false);
		return 0;
	}
	bool enable_button = true;
	ItemAssociatedData *ptr = reinterpret_cast<ItemAssociatedData*>(this->drivers_list.GetItemData(i));
	ItemAssociatedData::ItemState const item_state0 = ptr->getItemState();
	i = this->drivers_list.GetNextItem(i, LVIS_SELECTED);
	while (i != -1) {
		ptr = reinterpret_cast<ItemAssociatedData*>(this->drivers_list.GetItemData(i));
		ItemAssociatedData::ItemState item_state = ptr->getItemState();
		if (item_state != item_state0) {
			enable_button = false;
			break;
		}
		i = this->drivers_list.GetNextItem(i, LVIS_SELECTED);
	}
	if (item_state0 == ItemAssociatedData::ADDED_BEFORE) {
		this->remove_button.SetWindowText(L"削除する(&D)");
	} else if (item_state0 == ItemAssociatedData::ADDED_BEFORE_REMOVE) {
		this->remove_button.SetWindowText(L"削除の取り消し(&D)");
	} else {
		this->remove_button.SetWindowText(L"一覧から除外(&D)");
	}
	this->remove_button.EnableWindow(enable_button);
	this->property_button.EnableWindow(enable_button);

	return 0;
}

