#include "WorkManager.h"

#include "SourceSetting.h"
#include "SlipstreamSetting.h"
#include "OutputSetting.h"
#include "WorkStage.h"
#include "WorkingDirectoryPrepare.h"
#include "ServicePackSlipStream.h"
#include "DriverSlipStream.h"
#include "Driver32CabUpdater.h"
#include "SourceSizeReducer.h"
#include "TemporaryDirectory.h"
#include "ISOImageCreator.h"
#include "DiscRecorder.h"
#include "USBSetup.h"

extern void deleteDirectory(boost::filesystem::wpath const &dir_path);
extern void removeFileForcibly(boost::filesystem::wpath const &file_path);
extern void moveFileOrDirUsingWildCards(boost::filesystem::wpath const &src_path,
                                        boost::filesystem::wpath const &dst_dir_path);
extern void copyDirectory(boost::filesystem::wpath const &source_dir,
                          boost::filesystem::wpath const &destination_dir);

namespace fs {
	using namespace boost::filesystem;
};

/**
 * ċAIɃfBNgc[폜
 */
void deleteDirectory(boost::filesystem::wpath const &dir_path)
{
	MyLib::toFileWritable(dir_path);
	fs::wdirectory_iterator it_end;
	for (fs::wdirectory_iterator it(dir_path); it != it_end; ++it) {
		fs::wpath p = it->path();
		if (fs::is_directory(p)) {
			deleteDirectory(p);
		} else {
			removeFileForcibly(p);
		}
	}
	::RemoveDirectory(dir_path.string().c_str());
	//fs::remove(dir_path);
}

void removeFileForcibly(boost::filesystem::wpath const &file_path)
{
	MyLib::toFileWritable(file_path);
	//boost::filesystem::remove(file_path);
	::DeleteFile(file_path.string().c_str());
}

void moveFileOrDirUsingWildCards(boost::filesystem::wpath const &src_path,
                                 boost::filesystem::wpath const &dst_dir_path)
{
	assert(fs::is_directory(dst_dir_path));

	CFindFile finder;
	if (finder.FindFile(src_path.string().c_str())) {
		do {
			if (!finder.IsDots()) {
				TCHAR buf[MAX_PATH];
				finder.GetFilePath(buf, MAX_PATH);
				fs::wpath src = fs::wpath(buf);
				fs::wpath dst = dst_dir_path/src.filename();
				fs::rename(src, dst);
			}
		} while (finder.FindNextFile());
	}
}

void copyDirectory(boost::filesystem::wpath const &source_dir,
                   boost::filesystem::wpath const &destination_dir)
{
	fs::wdirectory_iterator it_end;
	for (fs::wdirectory_iterator it(source_dir); it != it_end; ++it) {
		fs::wpath src = it->path();
		fs::wpath dst = destination_dir/src.filename();
		if (fs::is_directory(src)) {
			//tH_Rs[ɍ쐬
			fs::create_directory(dst);
			//ċAĂяo
			copyDirectory(src, dst);
		} else {
			std::wstring const src_path_str = src.string();
			std::wstring const dest_path_str = dst.string();
			MyLib::toFileWritable(dst);
			if (!::CopyFile(src_path_str.c_str(), dest_path_str.c_str(), FALSE)) {
				throw MY_WIN32_FILE_TRANSFER_EXCEPTION_LAST_ERROR(src_path_str, dest_path_str);
			}
		}
	}
}

class TemporaryISOImageRemove
	: public WorkStage
{
protected:
	boost::filesystem::wpath iso_image_path_;

public:
	TemporaryISOImageRemove(boost::filesystem::wpath const &iso_image_path)
		: iso_image_path_(iso_image_path)
	{
	}

	virtual void execute()
	{
		boost::filesystem::remove(this->iso_image_path_);
	}
};

/////////////////////////////////////////////////////////////////////////////

WorkManager::WorkManager(SourceSetting const *source_setting,
                         SlipstreamSetting const *slipstream_setting,
                         OutputSetting const *output_setting)
{
	std::wstring source_path;
	type_info const &t_info = typeid(*source_setting);
	if (t_info == typeid(DirectorySourceSetting) || t_info == typeid(DiscSourceSetting)) {
		DirectorySourceSetting const *ptr = static_cast<DirectorySourceSetting const*>(source_setting);
		source_path = ptr->getSourcePath().string();
	}
	std::wstring const work_dir_path0 = slipstream_setting->getWorkDirPath().string();
	bool const is_work_dir_path_empty = work_dir_path0.empty();
	fs::wpath const work_dir_path =
		((is_work_dir_path_empty) ? source_path : work_dir_path0);

	enum {
		WORK_DIR_PREPARE,
		SP_SLIPSTREAM,
		DRIVER_SLIPSTREAM,
		DRIVER32_CAB,
		SOURCE_SIZE_REDUCER,
		ISO_IMAGE_CREATOR,
		DISC_RECORDER,
		TEMP_ISO_IMAGE_REMOVE,
		USB_SETUP,
		MAX_WORK_STAGE_COUNT
	};
	std::auto_ptr<WorkStage> work_stages[MAX_WORK_STAGE_COUNT];

	// ƃfBNgɈꎞfBNg쐬
	// ̈ꎞfBNgɃCXg[\[XꎞIɊi[
	TemporaryDirectoryCloseBehavior *ptr001;
	if (slipstream_setting->createsTempWorkDir()) {
		ptr001 = new TemporaryDirectoryCloseDelete();
	} else {
		ptr001 = new TemporaryDirectoryCloseUndoMovedFiles(work_dir_path);
	}
	this->temp_work_dir_close_behavior_.reset(ptr001);
	this->temp_work_dir_.reset(
		new TemporaryDirectory(work_dir_path,
		                       this->temp_work_dir_close_behavior_.get()));
	fs::wpath const temp_work_dir_path = this->temp_work_dir_->getDirectoryPath();
	// copy files if work directory name is not empty
	if (!is_work_dir_path_empty) {
		work_stages[WORK_DIR_PREPARE].reset(
			new WorkingDirectoryPrepareWithFileCopy(source_path,
			                                        temp_work_dir_path));
	} else {
		work_stages[WORK_DIR_PREPARE].reset(
			new WorkingDirectoryPrepare(source_path, temp_work_dir_path));
	}


	uint32_t const os_version = source_setting->getOSVersion();
	std::wstring const i386_subdir_name = source_setting->getI386SubdirName();

	// SP
	fs::wpath const sp_file_path = slipstream_setting->getSPFileName();
	if (!sp_file_path.string().empty()) {
		if (!fs::exists(sp_file_path) || !fs::is_regular_file(sp_file_path)) {
			//throw
		}
		work_stages[SP_SLIPSTREAM].reset(
			new ServicePackSlipStream(sp_file_path, os_version, i386_subdir_name, work_dir_path, temp_work_dir_path));
	}

	int const architecture_id = source_setting->getArchitectureID();
	// Driver
	std::vector<DriverSetupData> const add_drivers = slipstream_setting->getAddDriverCollection();
	DriverFileDataMap const leave_drivers = slipstream_setting->getLeaveDrivers();
	DriverFileDataMap const del_drivers = slipstream_setting->getDelDrivers();
	DriverServiceData const leave_services = slipstream_setting->getLeaveServices();
	DriverServiceData const del_services = slipstream_setting->getDelServices();
	bool const adapts_big_drive = slipstream_setting->usesBigdriveW2K();
	bool const recompresses_driver_cab = slipstream_setting->executesDriverCabRecompress();
	wchar_t const reserved_drive = output_setting->getReservedDrive();
	if (!add_drivers.empty() || !del_drivers.empty() || adapts_big_drive || recompresses_driver_cab) {
		work_stages[DRIVER_SLIPSTREAM].reset(new DriverSlipStream(add_drivers,
		                                                          leave_drivers,
		                                                          del_drivers,
		                                                          leave_services,
		                                                          del_services,
		                                                          reserved_drive,
		                                                          adapts_big_drive,
		                                                          recompresses_driver_cab,
		                                                          temp_work_dir_path,
		                                                          i386_subdir_name,
		                                                          architecture_id,
		                                                          os_version));
	}

	// Driver32.cab
	if (architecture_id == PROCESSOR_ARCHITECTURE_AMD64) {
		work_stages[DRIVER32_CAB].reset(new Driver32CabUpdater(temp_work_dir_path, i386_subdir_name));
	}

	// Reduce source size
	if (slipstream_setting->isCleanInstallOnly()) {
		work_stages[SOURCE_SIZE_REDUCER].reset(
			new SourceSizeReducer(temp_work_dir_path, i386_subdir_name, work_dir_path, os_version));
	}

	// ISOC[W
	bool is_temp_iso_image = false;
	std::wstring iso_image_path = output_setting->getISOImagePath().string();
	std::wstring const record_device_str = output_setting->getRecordDevice();
	// ISOC[W̖OȂCDĂꍇ
	if (!record_device_str.empty() && iso_image_path.empty()) {
		// ꎞt@CISOC[W쐬
		iso_image_path = MyLib::createTemporaryFilePath(work_dir_path).string();
		is_temp_iso_image = true;
	}
	//fs::wpath bin_dir_path;
	//{
	//	TCHAR mod_dir_buf[MAX_PATH];
	//	::GetModuleFileName(NULL, mod_dir_buf, MAX_PATH);
	//	fs::wpath module_dir_path = mod_dir_buf;
	//	bin_dir_path = module_dir_path.remove_filename();
	//	bin_dir_path /= L"bin";
	//}
	if (!iso_image_path.empty()) {
		// {[x
		std::wstring const volume_label = output_setting->getVolumeLabel();
		work_stages[ISO_IMAGE_CREATOR].reset(
			new ISOImageCreator(iso_image_path, temp_work_dir_path, source_path, i386_subdir_name, volume_label));
	}

	// CDĂ
	if (!record_device_str.empty()) {
		int record_speed = output_setting->getRecordSpeed();
		bool erases_rw_media = output_setting->erasesMedia();
		bool ejects_media = output_setting->ejectsMedia();
		work_stages[DISC_RECORDER].reset(
			new DiscRecorder(record_device_str, record_speed, erases_rw_media, ejects_media, iso_image_path));
	}

	// USBEDiskToDisk
	wchar_t const usb_target_drive = output_setting->getUSBTargetDrive();
	wchar_t const usb_setup_drive = output_setting->getUSBSetupDrive();
	if (::iswalpha(usb_target_drive) && ::iswalpha(usb_setup_drive)) {
		work_stages[USB_SETUP].reset(
			new USBSetup(temp_work_dir_path, i386_subdir_name, os_version, usb_target_drive, usb_setup_drive));
	}

	// ꎞISOC[W폜
	if (is_temp_iso_image) {
		work_stages[TEMP_ISO_IMAGE_REMOVE].reset(
			new TemporaryISOImageRemove(iso_image_path));
	}

	// OŜ
	this->work_stages_.reserve(MAX_WORK_STAGE_COUNT);
	for (int i = 0; i < MAX_WORK_STAGE_COUNT; ++i) {
		if (work_stages[i].get() != NULL) {
			this->work_stages_.push_back(work_stages[i].release());
		}
	}
}

WorkManager::~WorkManager()
{
	for (std::vector<WorkStage*>::iterator it = this->work_stages_.begin(); it != this->work_stages_.end(); ++it) {
		delete *it;
	}
}

void WorkManager::execute()
{
	std::for_each(this->work_stages_.begin(), this->work_stages_.end(), std::mem_fun(&WorkStage::execute));
}

void WorkManager::addEventListener(MyLib::EventListener<std::wstring> &event_listener)
{
	for (std::vector<WorkStage*>::iterator it = this->work_stages_.begin(); it != this->work_stages_.end(); ++it) {
		(*it)->addEventListener(&event_listener);
	}
}
